/* toon.c - functions for drawing cartoons on the root window - used by xpenguins
 * Copyright (C) 1999, 2000  Robin Hogan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Basic X-windows includes */
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <signal.h>
#include <limits.h>

/* Handle `virtual' window managers - courtesy of Andreas Stolcke */
#include "vroot.h"

#include "toon.h"

#define max(a,b) ( ( (a) > (b) ) ? (a) : (b) )

typedef int ErrorHandler();

Display *display;
int screen = 0;
Window root;
int display_width, display_height;
GC draw_toonGC;
Pixel black, white;
Region windows=NULL;
ToonData *toon_data;
int error_value = 0, edge_block=0;
/* If solid_popups is set to 0 then the toons fall behind `popup' windows.
 * This includes the KDE panel */
int solid_popups=1;
int toon_signal=0;
char toon_error_message[TOON_MESSAGE_LENGTH]="";
Toon *toons_to_erase=NULL;
int ntoons_to_erase=0;

/*** FUNCTIONS ***/

int ToonSignal() {
   return toon_signal;
}

void ToonSignalHandler(int sig) {
   toon_signal=sig;
   return;
}

void ToonExitGracefully(int sig) {
   ToonConfigure(TOON_NOCATCHSIGNALS);
   CloseToonDisplay();
   exit(0);
}

/*
void ToonExit(int sig) {
   exit(0);
}
*/
char * ToonErrorMessage() {
   return toon_error_message;
}

/* InitToonDisplay - open display, set GC etc. */
Display *OpenToonDisplay(char *display_name) {
   XGCValues gc_values;

   display=XOpenDisplay(display_name);
   if (display == NULL) {
      if (display_name == NULL && getenv("DISPLAY") == NULL)
         strncpy(toon_error_message,"DISPLAY environment variable not set",
               TOON_MESSAGE_LENGTH);
      else
         strncpy(toon_error_message,"cannot connect to X-server",
               TOON_MESSAGE_LENGTH);
      return(NULL);
   }
   screen = DefaultScreen(display);
   root = RootWindow(display, screen);
   black = BlackPixel(display, screen);
   white = WhitePixel(display, screen);
   display_width = DisplayWidth(display, screen);
   display_height = DisplayHeight(display, screen);

/* Set Graphics Context */
   gc_values.function = GXcopy;
   gc_values.graphics_exposures= False;
   gc_values.fill_style = FillTiled;
   draw_toonGC = XCreateGC(display,root,
      GCFunction | GCFillStyle | GCGraphicsExposures,&gc_values);

/* Regions */
   windows = XCreateRegion();

/* Notify if the root window changes */
   XSelectInput(display, root, SubstructureNotifyMask);

   return display;
}

int ToonConfigure(long int code) {
   if (code & TOON_EDGEBLOCK)
      edge_block=1;
   else if (code & TOON_NOEDGEBLOCK)
      edge_block=0;
   if (code & TOON_SOLIDPOPUPS)
      solid_popups=1;
   else if (code & TOON_NOSOLIDPOPUPS)
      solid_popups=0;
   if (code & TOON_CATCHSIGNALS) {
      signal(SIGINT, ToonSignalHandler);
      signal(SIGTERM, ToonSignalHandler);
      signal(SIGHUP, ToonSignalHandler);
   }
   else if (code & TOON_EXITGRACEFULLY) {
      signal(SIGINT, ToonExitGracefully);
      signal(SIGTERM, ToonExitGracefully);
      signal(SIGHUP, ToonExitGracefully);
   }
   else if (code & TOON_NOCATCHSIGNALS) {
      signal(SIGINT, SIG_DFL);
      signal(SIGTERM, SIG_DFL);
      signal(SIGHUP, SIG_DFL);
/*      signal(SIGINT, ToonExit);
      signal(SIGTERM, ToonExit);
      signal(SIGHUP, ToonExit);
*/
   }
   return 0;
}

int InstallToonData(ToonData *data, int n) {
   int i, status;
   for (i=0;i<n;i++) {
      (data+i)->attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions
				 | XpmExactColors | XpmCloseness);
      (data+i)->attributes.exactColors=False;
      (data+i)->attributes.closeness=40000;
      status = XpmCreatePixmapFromData(display, root, (data+i)->image,
         &((data+i)->pixmap), &((data+i)->mask), &((data+i)->attributes));
   }
   toon_data=data;
   return status;
}

int DrawToons(Toon *toon, int n) {
   int width,height,i;
   Toon *t;

   for (i=0;i<n;i++) {
      t=toon+i;
      width=toon_data[t->type].width;
      height=toon_data[t->type].height;

      XSetClipOrigin(display, draw_toonGC,
         t->x-width*t->frame, t->y-height*t->direction); 
      XSetClipMask(display, draw_toonGC, 
         toon_data[t->type].mask);   
      XCopyArea(display,
         toon_data[t->type].pixmap,
         root,draw_toonGC,width*t->frame,height*t->direction,
         width,height,t->x,t->y);
      XSetClipMask(display, draw_toonGC, None);
      t->x_map = t->x;
      t->y_map = t->y;
      t->width_map = width;
      t->height_map = height;
   }
   return 0;
}

int EraseToons(Toon *toon,int n) {
   int i;
   Toon *t;
   for (i=0;i<n;i++) {
      t=toon+i;
      XClearArea(display, root, t->x_map, t->y_map,
         t->width_map, t->height_map, False);
   }
   return 0;
}

int ToonBlocked(Toon *toon, int direction) {
   if (edge_block) {
      switch (direction) {
         case TOON_LEFT:
            if ((toon->x) <= 0) return 1;
            break;
         case TOON_RIGHT:
            if ((toon->x)+(toon_data[toon->type].width) >= display_width)
               return 1;
            break;
         case TOON_UP:
            if ((toon->y) <= 0) return 1;
            break;
         case TOON_DOWN:
            if ((toon->y)+(toon_data[toon->type].height) >= display_height)
               return 1;
            break;
      }
   }
   switch (direction) {
      case TOON_HERE:
         return (XRectInRegion(windows,toon->x,toon->y,
               toon_data[toon->type].width,toon_data[toon->type].height) 
               != RectangleOut);
      case TOON_LEFT:
         return (XRectInRegion(windows,toon->x-1,toon->y,
               1,toon_data[toon->type].height) != RectangleOut);
      case TOON_RIGHT:
         return (XRectInRegion(windows,toon->x+toon_data[toon->type].width,
               toon->y,1,toon_data[toon->type].height) != RectangleOut);
      case TOON_UP:
         return (XRectInRegion(windows,toon->x,toon->y-1,
               toon_data[toon->type].width,1) != RectangleOut);
      case TOON_DOWN:
         return (XRectInRegion(windows,toon->x,
               toon->y+toon_data[toon->type].height, toon_data[toon->type].width,
               1) != RectangleOut);
      default:
         return -1;
   }
}

int AdvanceToon(Toon *toon, int mode) {
   int newx, newy;
   int new_zone;
   int width, height;
   int move_ahead = 1;
   int result=TOON_OK;

   if (mode == TOON_STILL) move_ahead = 0;

   width=toon_data[toon->type].width;
   height=toon_data[toon->type].height;

   newx = toon->x + toon->u;
   newy = toon->y + toon->v;

   if (edge_block) {
      if (newx < 0) {
         newx=0;
         result=TOON_PARTIALMOVE;
      }
      else if (newx + toon_data[toon->type].width > display_width) {
         newx=display_width-toon_data[toon->type].width;
         result=TOON_PARTIALMOVE;
      }
      if (newy < 0) {
         newy=0;
         result=TOON_PARTIALMOVE;
      }
      else if (newy + toon_data[toon->type].height > display_height) {
         newy=display_height-toon_data[toon->type].height;
         result=TOON_PARTIALMOVE;
      }
   }

   /* Is new toon location fully/partially filled with windows? */
   new_zone = XRectInRegion(windows,newx,newy,width,height);
   if (new_zone != RectangleOut && mode == TOON_MOVE) {
      int tryx, tryy, step=1;
      result=TOON_BLOCKED;
      move_ahead=0;
      /* How far can we move the toon? */
      if ( abs(toon->v) < abs(toon->u) ) {
         if (newx>toon->x) step=-1;
         for (tryx=newx+step;tryx!=(toon->x);tryx=tryx+step) {
            tryy=toon->y+((tryx-toon->x)*(toon->v))/(toon->u);
            if (XRectInRegion(windows,tryx,tryy,width,height) == RectangleOut) {
               newx=tryx;
               newy=tryy;
               result=TOON_PARTIALMOVE;
               move_ahead=1;
               break;
            }
         }
      }
      else {
         if (newy>toon->y) step=-1;
         for (tryy=newy+step;tryy!=(toon->y);tryy=tryy+step) {
            tryx=toon->x+((tryy-toon->y)*(toon->u))/(toon->v);
            if (XRectInRegion(windows,tryx,tryy,width,height) == RectangleOut) {
               newx=tryx;
               newy=tryy;
               result=TOON_PARTIALMOVE;
               move_ahead=1;
               break;
            }
         }
      }
   }
   if (move_ahead) {
      toon->x=newx;
      toon->y=newy;
      if ( (++(toon->frame)) >= toon_data[toon->type].nframes) {
         toon->frame = 0;
         if ( (toon_data[toon->type].conf) & TOON_NOCYCLE) 
            toon->active = 0;
      }
   }
   return result;
}

void FlushToons() {
   XFlush(display);
   return;
}

int ToonError(Display *display, XErrorEvent *error) {
    error_value = error->error_code;
    return 0;
}

int ToonDisplayWidth() {
   return display_width;
}

int ToonDisplayHeight() {
   return display_height;
}

/* Build up an X-region corresponding to the location of the windows 
 * that we don't want our toons to enter */
int ToonLocateWindows() {
   Window *children;
   unsigned int nchildren;
   Window dummy;
   XWindowAttributes attributes;
   int wx;
   XRectangle window_rect;
   int x, y;
   unsigned int height, width;
   unsigned int depth, border_width;

   XSetErrorHandler(ToonError);

   /* Rebuild window region */
   XDestroyRegion(windows);
   windows = XCreateRegion();

   /* Get children of root */
   XQueryTree(display, root, &dummy, &dummy, &children, &nchildren);

   /* Add windows to region */
   for (wx=0; wx<nchildren; wx++) {
      error_value = 0;
      XGetWindowAttributes(display, children[wx], &attributes);
      if (error_value) continue;

      /* Popup? */
      if (!solid_popups && attributes.save_under) continue;

      if (attributes.map_state == IsViewable) {
         XGetGeometry(display, children[wx], &dummy, &x, &y,
            &width, &height, &border_width, &depth);
         if (error_value) continue;

         /* Entirely offscreen? */
         if (x > display_width) continue;
         if (y > display_height) continue;
         if (y <= 0) continue;

         /* Geometry of the window, borders inclusive */
         x = x-border_width;
         y = y-border_width;
         height = height + 2*border_width;
         width = width + 2*border_width;

         /* Entirely offscreen? */
         if ((x + width) < 0) continue;

         if (y < 0) {
            height = height + y;
            y = 0;
         }

         /* If window partly on screen */
         if (x < 0) {
            width = width + x - border_width;
            x = 0;
         }

         window_rect.x = x;
         window_rect.y = y;
         window_rect.height = height + 2*border_width;
         window_rect.width = width + 2*border_width;

         /* The area of the windows themselves */
         XUnionRectWithRegion(&window_rect, windows, windows);
      }
   }
   XFree((char *)children);
   XSetErrorHandler((ErrorHandler *)NULL);
   if (XEventsQueued(display, QueuedAlready)) {
      return 1;
   }
   return 0;
}

int SetToonType(Toon *toon, int type, int direction) {
   toon->type = type;
   toon->direction = direction;
   toon->frame = 0;
   toon->active = 1;
   return 0;
}

int SetToonPosition(Toon *toon, int x, int y) {
   toon->x=x;
   toon->y=y;
   return 0;
}

int SetToonVelocity(Toon *toon, int u, int v) {
   toon->u=u;
   toon->v=v;
   return 0;
}

int CloseToonDisplay() {
   XDestroyRegion(windows);
   XClearWindow(display,root);
   XCloseDisplay(display);
   return 0;
}

int ToonSleep(unsigned long usecs) {
   struct timeval t;
   t.tv_usec = usecs%(unsigned long)1000000;
   t.tv_sec = usecs/(unsigned long)1000000;
   select(0, (void *)0, (void *)0, (void *)0, &t);
   return 0;
}

int ToonWindowsMoved() {
   XEvent event;
   int windows_moved=0;
   while (XPending(display)) {
      XNextEvent(display, &event);
      switch (event.type) {
         case ConfigureNotify:case MapNotify:case UnmapNotify:
            windows_moved=1;
            break;
      }
   }
   if (windows_moved) {
      ToonLocateWindows();
   }
   return windows_moved;
}

